#include <propeller.h>
#include <stdint.h>
#include <stdio.h>

#include "lvds.h"
#include "font.h"

__asm {
            org 0
entry       rdlong pin, ptra[0]
            setxfrq pix_freq // set streamer frequency
            // set r1 to 0110 0000 1 [pin] 0
            mov r1, #0x6
            shl r1, #5
            add r1, #1
            shl r1, #6
            add r1, pin
            shl r1, #17 // total shift by 32

            add r1, #7 // set duration

            // set up the 4 pins for output
            dirh pin
            add pin, #1
            dirh pin
            add pin, #1
            dirh pin
            add pin, #1
            dirh pin

            // setup a trigger pin for debugging
            dirh #0

            mov fb, ptra
            add fb, #12
            xinit r1, pixel_buf

            dirh #12
            outh #12

// start of frame
frame       drvnot #0
            rdfast #0, fb // setup the RAM fifo at the start of the screen

// start of v sync
//             mov r3, twv             // start vsync with:
//             mov vs, #0              //      vs = 0
//             mov de_v, #0            //      de override to 0
//             mov pix_rtn, #blk_pixel //      blank pixels

// vsync       call #line
//             djnz r3, #vsync

// // start of v back porch
//             mov r3, tvb             // start back porch with:
//             mov vs, vs_bit          //      vs = 1
//             mov de_v, #0            //      de override to 0
//             mov pix_rtn, #blk_pixel //      blank pixels

// vbp         call #line
//             djnz r3, #vbp

// start of v active area
            mov r3, tva             // start back porch with:
            mov vs, vs_bit          //      vs = 1
            mov de_v, de_bit        //      allow line to control de
            mov pix_rtn, #pixel     //      real pixels
            mov ln_cnt, #0          // reset the line counts

vac         call #line
            add ln_cnt, #1
            djnz r3, #vac

// start of v front porch
            mov r3, tvf             // start back porch with:
            mov vs, vs_bit          //      vs = 1
            mov de_v, #0            //      de override to 0
            mov pix_rtn, #blk_pixel //      blank pixels

vfp         call #line
            djnz r3, #vfp

            jmp #frame // return back to start of frame

/* Draw line sub routine */
line
// // start of h sync
//             mov r4, twh     // start of line is twh number of pixels with:
//             mov hs, #0      //      hs = 0
//             mov de, #0      //      data enable = 0

// hsync       call #blk_pixel // draw the pixel
//             djnz r4, #hsync // continue hsync until counter is 0

// // start of back porch
//             mov r4, thb     // now the backporch, with:
//             mov hs, hs_bit  //      hs = 1
//             mov de, #0      //      data enable = 0

// hbp         call #blk_pixel // loop pixels until counter is 0
//             djnz r4, #hbp

// start of active area
            mov r4, tha     // now the actual data, with
            mov hs, hs_bit  //      hs = 1
            mov de, de_v    //      data enable = 1 (only when in active vertical area)
            mov h_cnt, #0

hac         call pix_rtn    // loop pixels until counter is 0
            add h_cnt, #1
            djnz r4, #hac

// start of front porch
            mov r4, thf     // now the front porch, with:
            mov hs, hs_bit  //      hs = 1
            mov de, #0      //      data enable = 0

hfp         call #blk_pixel // loop pixels until counter is 0
            djnz r4, #hfp
            ret // done with line, return

/* Draw pixel sub routine */
pixel       mov pixel_buf, pix_clk
            or pixel_buf, hs    // put the value of hs into the pixel buffer
            or pixel_buf, de    // put the value of de into the pixel buffer
            or pixel_buf, vs    // put the value of vs into the pixel buffer

            // we have room for 9 operations to figure out the pixel data.
            mov r7, h_cnt
            and r7, #7  wz      // pixel offset into current byte of the frame buffer
    if_z    rfbyte r8           // r8 = a byte from the pixel buffer

            mov r2, r8          // r2 isn't used, can re-use it
            shr r2, r7          // shift the byte by the offset into the current byte
            and r2, #1 wz       // get the first bit, z = result == 0

    if_z    or pixel_buf, bg_bits // if pixel is 0, set color bits to the background color
    if_nz   or pixel_buf, fg_bits // if pixel is 1, set color bits to the foreground color

            //or pixel_buf, fg_bits

            xcont r1, pixel_buf
            ret

/* Draw blank pixel sub routine */
blk_pixel   mov pixel_buf, pix_clk
            or pixel_buf, hs    // put the value of hs into the pixel buffer
            or pixel_buf, de    // put the value of de into the pixel buffer
            or pixel_buf, vs    // put the value of vs into the pixel buffer

            xcont r1, pixel_buf
            ret

r0          long    0 // some general registers
r1          long    0
r2          long    0
r3          long    0
r4          long    0
r5          long    0 // stores the tile to be drawing from
r6          long    0
r7          long    0
r8          long    0

h_cnt       long    0 // pixel count along line
ln_cnt      long    0 // line count

pin         long    0
pixel_buf   long    0
hs          long    1<<11 // hs starts high
vs          long    1<<7  // vs starts high
de          long    0
de_v        long    0 // data enable control for whole lines
pix_rtn     long    0 // the pixel routing to call. should be either #pixel or #blk_pixel

fb          long    0 // stores the hub address of the start of the screen buffer

pix_freq    long    0x16000000 // multipler (out of 0x80000000) to set the pixel frequency.

pix_clk     long    0x1100011 // 7 nibbles form a pixel, with bit 0 of each nibble storing the clock value
// hs_bit      long    1<<11   // horizontal sync bit
// vs_bit      long    1<<7    // vertical sync bit
hs_bit      long    0   // horizontal sync bit
vs_bit      long    0    // vertical sync bit
de_bit      long    1<<3    // data enable bit

white       long    0xEEEE666 // these are all the color bits
fg_bits     long    0xEEEE666
bg_bits     long    0x8888444

// timing parameters
// twh         long    1      // h sync pixels
// thb         long    160      // h back porch pixels
tha         long    1024   // h active area pixels
thf         long    160      // h front porch pixels

// twv         long    1       // v sync lines
// tvb         long    23     // v back porch lines
tva         long    600     // v active area lines
tvf         long    10       // v front porch lines
}


lvds_t *LVDS(lvds_t *l) {
    l->start = lvds_t_start;
    l->_char = lvds_t_char;
    l->str = lvds_t_str;
    return l;
}

int32_t lvds_t_start(lvds_t *l, uint8_t pin) {
    l->p = pin;
    return _cognew(&entry, l);
}

void lvds_t_char(lvds_t *l, int32_t x, int32_t y, char c) {
    c = c - 0x20; // don't have first 0x20 characters in the ascii table, they are all blanks

    for (int i = 0; i < 16; i++) {
        uint32_t c_byte = font8x16[16*c + i] << 24;
        __asm {
            rev c_byte
        };
        l->fb[x + (16*y+i)*LVDS_TEXT_COLS] = c_byte & 0xff;
    }
}

void lvds_t_str(struct lvds_t *l, int32_t x, int32_t y, const char *c) {
    while(*c) {
        l->_char(l, x++, y, *c);
        c++;
    }
}

